require "util"

require("config")
require("types.entityTypes")
require("control.regularPole")
require("control.powerBox")
require("control.highVoltage")

function prepareMod(event)
	global.randomGen			= global.randomGen or game.create_random_generator()
	global.regularPoles			= global.regularPoles or {}
	global.regularPolesDirty	= global.regularPolesDirty or {}
	global.highPoles			= global.highPoles or {}
	global.highPolesDirty		= global.highPolesDirty or {}
	global.powerBoxes			= global.powerBoxes or {}
	global.powerBoxesDirty		= global.powerBoxesDirty or {}
	global.transformers			= global.transformers or {}
	global.transformersDirty	= global.transformersDirty or {}
	global.indicatorBeams		= global.indicatorBeams or {}
	global.textHelpers			= global.textHelpers or {}
end

function findRegularPoleByEntity(pole)
	local hash = hash_entity(pole);
	local pole = global.regularPoles[hash];
	return pole;
end

function findHighPoleByEntity(pole)
	local hash = hash_entity(pole);
	local pole = global.highPoles[hash];
	return pole;
end

function findOrAddHighPoleByEntity(ent)
	local hash = hash_entity(ent);
	local pole = global.highPoles[hash];
	
	if pole then
		return pole
	end
	
	effectiveType = prototypes[ent.name]
	if effectiveType and effectiveType.type == "highPole" then	
		onBuiltEntity({created_entity=ent})
		pole = global.highPoles[hash];
		--debugOutputMessage("Rebuilt pole: " .. pole.hash);
	end
	
	return pole
end

function findPowerBoxByEntity(pole)
	local hash = hash_entity(pole);
	local box = global.powerBoxes[hash];
	return box;
end

function findTransformerByEntity(pole)
	local hash = hash_entity(pole);
	local pole = global.transformers[hash];
	return pole;
end

function makeRegularPoleDirty(pole)
	pole.dirtyCountDown = 2;
--	if not global.regularPolesDirty[pole.hash] then
--		debugOutputMessage("Now dirty: " .. pole.hash);
--	end
	global.regularPolesDirty[pole.hash] = pole.hash;
end

function makeRegularPoleNetworkDirty(pole)
	for k,v in pairs(pole.peers) do
		local peer = global.regularPoles[v]
		if peer then
			makeRegularPoleDirty(peer);
		end		
	end
	if pole.powerBox then
		local powerBox = global.powerBoxes[pole.powerBox];
		if powerBox then
			makePowerBoxDirty(powerBox);
		end
	end
end

function makeHighPoleDirty(pole)
	pole.dirtyCountDown = 2;
	global.highPolesDirty[pole.hash] = pole.hash;
end

function makeHighPoleNetworkDirty(pole)
	if pole.peerA then
		local peerAsPoleA = global.highPoles[pole.peerA];
		if peerAsPoleA then
			makeHighPoleDirty(peerAsPoleA);
		end		
	end
	if pole.peerB then
		local peerAsPoleB = global.highPoles[pole.peerB];
		if peerAsPoleB then
			makeHighPoleDirty(peerAsPoleB);
		end		
	end
	
	local associatedTransformer = pole.transformerDirect or pole.transformer
	
	--debugOutputMessage("High Pole made dirty: " .. pole.hash);

	if associatedTransformer then
		local trans = global.transformers[associatedTransformer];
		if trans then
			--debugOutputMessage("Transformer 1 made dirty: " .. trans.hash);
			makeTransformerDirty(trans);
			local other = trans.peerMajor or trans.peerMinor;
			trans.peerMajor = nil
			trans.peerMinor = nil
			if other then
				--debugOutputMessage("Transformer 2 made dirty: " .. other.hash);
				other.peerMajor = nil
				other.peerMinor = nil
				makeTransformerDirty(other);
			end
		end
	end
end

function makePowerBoxDirty(box)
	box.dirtyCountDown = 2;
	global.powerBoxesDirty[box.hash] = box.hash;
end

function makePowerBoxNetworkDirty(box)
	for k,v in pairs(box.regularPoles) do
		local pole = global.regularPoles[v]
		if pole then
			makeRegularPoleDirty(pole);
		end		
	end
end

function makeTransformerDirty(trans)
	local transformer = trans.transformer;
	
	--debugOutputMessage("Transformer made dirty: " .. trans.hash .. "(" .. debug.traceback() .. ")");
	
	if not transformer then
		return
	end
	if not transformer.valid then
		removeTransformerData(trans)
		return
	end
	
	local items = transformer.surface.find_entities_filtered{area = {{-10 + transformer.position.x, -10 + transformer.position.y}, {10 + transformer.position.x, 10 + transformer.position.y}}, type="electric-pole"}
	for k,v in pairs(items) do
		local proto = prototypes[v.name];
		
		if proto and proto.type == "highPole" then
			local pole = findHighPoleByEntity(v)
			if pole then
				makeHighPoleDirty(pole)
			end
		end
	end
	trans.dirtyCountDown = 2;
	global.transformersDirty[trans.hash] = trans.hash;
end

function handleSelection(player)
	local poles = global.regularPoles;
	local box = global.powerBoxes;
	local highPoles = global.highPoles;
	local transformers = global.transformers;
	
	local selected = player.selected;
	local objectAsPole = nil;
	local objectAsBox = nil;
	local objectAsHighPole = nil;
	local objectAsTransformer = nil;
	
	if selected and selected.valid then
		local hash = hash_entity(selected);
	
		objectAsPole = poles[hash]
		if objectAsPole then goto FoundObject end
		
		objectAsBox = box[hash]
		if objectAsBox then goto FoundObject end
		
		objectAsHighPole = highPoles[hash]
		if objectAsHighPole then goto FoundObject end
		
		objectAsTransformer = transformers[hash]
		if objectAsTransformer then goto FoundObject end
		::FoundObject::
	end
	
	if objectAsPole then
		makeRegularPoleDirty(objectAsPole);
		updateRegularPoleInfoPanel(player, objectAsPole);
		showInwardRegularPoleIndicators(objectAsPole, 1);
		showOutwardRegularPoleIndicators(objectAsPole, 1);
	else
		removeRegularPoleInfoPanel(player);
	end
	
	if objectAsBox then
		makePowerBoxDirty(objectAsBox);
		showOutwardPowerBoxIndicators(objectAsBox, 1);
	else
	end
	
	if objectAsHighPole then
		makeHighPoleDirty(objectAsHighPole);
		updateHighPoleInfoPanel(player, objectAsHighPole);
		showHighPoleIndicators(objectAsHighPole, 1);
	else
		removeHighPoleInfoPanel(player);
	end
	
	if objectAsTransformer then
		makeTransformerDirty(objectAsTransformer);
		updateTransformerInfoPanel(player, objectAsTransformer);
		showTransformerIndicators(objectAsTransformer, 1);
	else
		removeTransformerInfoPanel(player);
	end
end

function tick(event)
	local regularPoles		= global.regularPoles;
	local regularPolesDirty	= global.regularPolesDirty;
	local highPoles			= global.highPoles;
	local highPolesDirty	= global.highPolesDirty;
	local powerBoxes		= global.powerBoxes;
	local powerBoxesDirty	= global.powerBoxesDirty;
	local transformers		= global.transformers;
	local transformersDirty	= global.transformersDirty;
	local beamHelpers		= global.indicatorBeams;
	
	local updateDirtyRegularPoles	= 0
	local updateDirtyHighPoles		= 0
	local updateDirtyPowerBox		= 0
	local updateDirtyTransformer	= 0
	local updateTransformers		= 0
	local updateBeamHelpers			= 0
	
	for k,v in pairs(game.players) do
		handleSelection(v);
	end
  
	for k,v in pairs(regularPolesDirty) do
		local pole = regularPoles[v];
		if not pole then
			regularPolesDirty[k] = nil;
			goto continue
		end
		if (pole.dirtyCountDown < 0) then
			regularPolesDirty[k] = nil;
			goto continue
		end
		pole.dirtyCountDown = pole.dirtyCountDown - 1;
		updateRegularPole(pole);
		updateDirtyRegularPoles = updateDirtyRegularPoles + 1
		::continue::
	end
  
	for k,v in pairs(highPolesDirty) do
		local pole = highPoles[v];
		if not pole then
			highPolesDirty[k] = nil;
			goto continue
		end
		if (pole.dirtyCountDown < 0) then
			highPolesDirty[k] = nil;
			goto continue
		end
		pole.dirtyCountDown = pole.dirtyCountDown - 1;
		updateHighPole(pole);
		updateDirtyHighPoles = updateDirtyHighPoles + 1
		::continue::
	end
  
	for k,v in pairs(powerBoxesDirty) do
		local box = powerBoxes[v];
		if not box then
			powerBoxesDirty[k] = nil;
			goto continue
		end
		if (box.dirtyCountDown < 0) then
			powerBoxesDirty[k] = nil;
			goto continue
		end
		box.dirtyCountDown = box.dirtyCountDown - 1;
		updatePowerBox(box);
		updateDirtyPowerBox = updateDirtyPowerBox + 1
		::continue::
	end
  
	for k,v in pairs(transformersDirty) do
		local trans = transformers[v];
		if not trans then
			transformersDirty[k] = nil;
			goto continue
		end
		if (trans.dirtyCountDown < 0) then
			transformersDirty[k] = nil;
			goto continue
		end
		trans.dirtyCountDown = trans.dirtyCountDown - 1;
		updateTransformer(trans);
		updateDirtyTransformer = updateDirtyTransformer + 1
		::continue::
	end
  
	--for k,v in pairs(transformers) do
	--	local trans = transformers[v];
	--	if not trans then
	--		transformersDirty[k] = nil;
	--		goto continue
	--	end
	--	updateTransformer(trans);
	--	::continue::
	--end
	
	for k,v in pairs(beamHelpers) do
		v.duration = v.duration - 1;
		if (not beamHelpers[k].beamEntity.valid) then
			beamHelpers[k] = nil;
			goto continue;
		end
		if (v.duration < 0) then
			beamHelpers[k].beamEntity.destroy();
			beamHelpers[k] = nil;
		end
		updateBeamHelpers = updateBeamHelpers + 1
		::continue::
	end
  
	for k,trans in pairs(transformers) do
		manageTransformer(trans);
		updateTransformers = updateTransformers + 1
	end
	
	--debugOutputMessage("Regular: " .. updateDirtyRegularPoles .. " - High: " .. updateDirtyHighPoles .. " - PowerBox: " .. updateDirtyPowerBox .. " - Transformers: " .. updateDirtyTransformer .. " - Connexions: " .. updateTransformers .. " - Helpers: " .. updateBeamHelpers);
end
	
function onBuiltEntity(event)
	local entity = event.created_entity
	local effectiveType = prototypes[entity.name];
	  
	if not effectiveType then
		return
	end
	  
	if effectiveType.type == "regularPole" then
		addRegularPole(entity, effectiveType)
	elseif effectiveType.type == "highPole" then
		addHighPole(entity, effectiveType)
	elseif effectiveType.type == "powerBox" then
		addPowerBox(entity, effectiveType)
	elseif effectiveType.type == "transformer" then
		addTransformer(entity, effectiveType)
	end
end
	
function onEntityDied(event)
	local entity = event.entity;
	
	if not entity.valid then
		return
	end
	
	local effectiveType = prototypes[entity.name];
	
	if not effectiveType then
		return
	end
	  
	if effectiveType.type == "regularPole" then
		local pole = findRegularPoleByEntity(entity);
		if pole then
			removeRegularPoleData(pole);
		end
	elseif effectiveType.type == "highPole" then
		local pole = findHighPoleByEntity(entity);
		if pole then
			removeHighPoleData(pole);
		end
	elseif effectiveType.type == "powerBox" then
		local box = findPowerBoxByEntity(entity);
		if box then
			removePowerBoxData(box);
		end
	elseif effectiveType.type == "transformer" then
		local trans = findTransformerByEntity(entity);
		if trans then
			removeTransformerData(trans);
		end
	end
end

function addBeamHelper(type, targetA, targetB, duration)
	if not targetA.valid or not targetB.valid then
		return
	end

	local superhash = hash_entity(targetA) .. "X" .. hash_entity(targetB) .. "X" .. type;
	local beam = global.indicatorBeams[superhash];
	
	if beam then
		beam.duration = math.max(beam.duration, duration);
		return;
	end
	
	beam = {
		beamEntity;
		duration = duration;
	};
	
	local surface = targetA.surface;
	beam.beamEntity = surface.create_entity({name = type, position = targetA.position, source = targetA, target = targetB});	
	
	global.indicatorBeams[superhash] = beam;
end

function addTextHelper(setup)
	if not setup.name then setup.name = "flying-text" end
	if not setup.text then setup.text = "error" end
	if not setup.color then setup.color = {r=1, g=.5, b=.5} end
	
	setup.position.x = setup.position.x - 0.5;
	setup.position.y = setup.position.y + 0.5;
	
	game.surfaces[1].create_entity(setup);
end

function debugOutputMessage(message)
	game.players[1].print(game.tick .. " " .. message);
end

function hash_entity(entity)
  return entity.position.x.."_"..entity.position.y
end

function gameLogicDistance(x0, y0, x1, y1)
	return math.max(math.abs(x0 - x1), math.abs(y0 - y1));
end

script.on_event(defines.events.on_built_entity, onBuiltEntity)
script.on_event(defines.events.on_robot_built_entity, onBuiltEntity)
script.on_event(defines.events.on_entity_died, onEntityDied)
script.on_event(defines.events.on_robot_pre_mined, onEntityDied)
script.on_event(defines.events.on_player_mined_entity, onEntityDied)
script.on_event(defines.events.on_tick, tick)
script.on_configuration_changed(prepareMod)
script.on_init(prepareMod)

