function addHighPole(entity, prototype)
	local existingPole = findHighPoleByEntity(entity);
	
	if existingPole then
		makeHighPoleNetworkDirty(existingPole);
		return;
	end
	
    local object = {
       pole = entity,
	   arrow = nil,
	   hash = hash_entity(entity),
	   calcDistance = 9999,
	   calcEfficiency = 0,
	   dirtyCountDown = 0,
	   peerA = nil,
	   peerB = nil,
	   enabled = true,
	   distanceQuality = 0.99,
	   transformer = nil,
	   transformerDirect = nil
    }
	
	global.highPoles[object.hash] = object;
	makeHighPoleDirty(global.highPoles[object.hash]);
	
	object.pole.disconnect_neighbour();
	
	--debugOutputMessage("Adding high pole data: " .. object.hash);
	
	return global.highPoles[object.hash];
end

function addTransformer(entity, prototype)
	local existingTransformer = findTransformerByEntity(entity);
	
	if existingTransformer then
		makeTransformerDirty(existingTransformer);
		return;
	end
	
    local object = {
       transformer = entity,
	   arrowDownA = nil,
	   arrowDownB = nil,
	   arrowDownC = nil,
	   arrowUpA = nil,
	   arrowUpB = nil,
	   arrowUpC = nil,
	   hash = hash_entity(entity),
	   pole = nil,
	   enabled = true,
	   peerMajor = nil,
	   peerMinor = nil,
	   effectiveEfficiency = 0,
	   interfaceDrain = nil,
	   interfaceOutput = nil,
	   calcDistance = 9999
    }
	
	object.interfaceDrain	= entity.surface.create_entity{name="transformer-interface-drain", position=entity.position, force=entity.force}
	object.interfaceOutput	= entity.surface.create_entity{name="transformer-interface-output", position=entity.position, force=entity.force}
	
	--debugOutputMessage("Adding transformer data: " .. object.hash);
	
	global.transformers[object.hash] = object;
	makeTransformerDirty(global.transformers[object.hash]);
	
	return global.highPoles[object.hash];
end

function removeHighPoleData(pole)
	--debugOutputMessage("Removing high pole data: " .. pole.hash);
	
	if pole.peerA then
		local peer = global.regularPoles[pole.PeerA]
		if peer then	
			peer.peerA = nil
			peer.peerB = nil
		end
	end
	if pole.peerB then
		local peer = global.regularPoles[pole.PeerB]
		if peer then	
			peer.peerA = nil
			peer.peerB = nil
		end
	end
	
	if pole.arrow and pole.arrow.valid then
		pole.arrow.destroy()
	end
	if pole.transformerDirect then
		local trans = global.transformers[pole.transformerDirect]
		if trans then
			trans.pole = nil
			if trans.peerMajor then
				trans.peerMajor.peerMinor = nil
				makeTransformerDirty(trans.peerMajor);
			end
			trans.peerMajor = nil
			trans.peerMinor = nil
			makeTransformerDirty(trans);
		end
	end
	
	makeHighPoleNetworkDirty(pole)
	
	global.highPoles[pole.hash] = nil
end

function removeTransformerData(trans)
	--debugOutputMessage("Removing transformer data: " .. trans.hash);
	
	if trans.arrow and trans.arrow.valid then
		trans.arrow.destroy()
	end
	
	if trans.transformer then
		makeTransformerDirty(trans)
	end
	
	trans.transformer = nil;
	
	if trans.interfaceDrain and trans.interfaceDrain.valid then
		trans.interfaceDrain.destroy();
	end
	if trans.interfaceOutput and trans.interfaceOutput.valid then
		trans.interfaceOutput.destroy();
	end
	
	global.transformers[trans.hash] = nil	
	
	if trans.arrowDownA and trans.arrowDownA.valid then
		trans.arrowDownA.destroy()
		trans.arrowDownA = nil
	end
	if trans.arrowDownB and trans.arrowDownB.valid then
		trans.arrowDownB.destroy()
		trans.arrowDownB = nil
	end
	if trans.arrowDownC and trans.arrowDownC.valid then
		trans.arrowDownC.destroy()
		trans.arrowDownC = nil
	end
	if trans.arrowUpA and trans.arrowUpA.valid then
		trans.arrowUpA.destroy()
		trans.arrowUpA = nil
	end
	if trans.arrowUpB and trans.arrowUpB.valid then
		trans.arrowUpB.destroy()
		trans.arrowUpB = nil
	end
	if trans.arrowUpC and trans.arrowUpC.valid then
		trans.arrowUpC.destroy()
		trans.arrowUpC = nil
	end
end

function updateHighPole(pole)
	--debugOutputMessage("Updating high pole: " .. pole.hash .. " - " .. pole.dirtyCountDown);
	
	if not pole.pole.valid then
		removeHighPoleData(pole)
		return
	end
	
	local electricConnected	= pole.pole.neighbours["copper"];
	local redConnected		= pole.pole.neighbours["red"];
	
	for k,v in pairs(electricConnected) do
		addTextHelper{position=pole.pole.position, text="(X)"}
		addTextHelper{position=v.position, text="(X)"}
	end
	
	pole.pole.disconnect_neighbour();
	
	local previousPeerA				= pole.peerA;
	local previousPeerB				= pole.peerB;
	local previousTransformer		= pole.transformer;
	local previousTransformerDirect	= pole.transformerDirect;
	local previousCalcDistance		= pole.calcDistance;
	local previousCalcEfficiency	= pole.calcEfficiency;
	
	local anyChanges		= false
	local redPeerCount		= 0
	local directTransformer	= nil
	local connectedPoles	= {}
	
	pole.peerA				= nil
	pole.peerB				= nil
	pole.transformer		= nil
	pole.calcDistance		= 9999999
	pole.transformerDirect	= nil
	pole.calcEfficiency		= 0;
	
	for k,v in pairs(redConnected) do
		local otherPole = findOrAddHighPoleByEntity(v)
		
		if redPeerCount < 2 then
			if otherPole then
				connectedPoles[otherPole.hash] = otherPole;
				redPeerCount		= redPeerCount + 1
				goto continue;
			end
			
			local otherTransformer = findTransformerByEntity(v)
			if otherTransformer then
				redPeerCount			= redPeerCount + 1
				pole.transformer		= otherTransformer.hash;
				pole.transformerDirect	= otherTransformer.hash;
				directTransformer		= otherTransformer;
				pole.calcDistance		= 0;
				pole.calcEfficiency		= 1;
				goto continue
			end
		end
		
		debugOutputMessage("Red disconnected: " .. redPeerCount);
		
		addTextHelper{position=pole.pole.position, text="(X)"}
		addTextHelper{position=v.position, text="(X)"}
		pole.pole.disconnect_neighbour{wire=defines.wire_type.red, target_entity=v};
		anyChanges = true
			
		::continue::
	end
	
	for k,otherPole in pairs(connectedPoles) do
		if not pole.peerA then
			pole.peerA			= otherPole.hash
		else
			pole.peerB			= otherPole.hash
		end
		if otherPole.transformer then
			if not pole.transformer or otherPole.transformer > pole.transformer then
				pole.transformer	= otherPole.transformer;
				pole.calcDistance	= otherPole.calcDistance + gameLogicDistance(pole.pole.position.x, pole.pole.position.y, otherPole.pole.position.x, otherPole.pole.position.y);
				pole.calcEfficiency	= otherPole.calcEfficiency * math.pow(.99995, pole.calcDistance);
			elseif otherPole.transformer == pole.transformer then
				pole.calcDistance	= math.min(pole.calcDistance, otherPole.calcDistance + gameLogicDistance(pole.pole.position.x, pole.pole.position.y, otherPole.pole.position.x, otherPole.pole.position.y));
				pole.calcEfficiency	= math.max(pole.calcEfficiency, otherPole.calcEfficiency * math.pow(.99995, pole.calcDistance));
			end
		end
		
		::continue::
	end
	
	if directTransformer then	
		directTransformer.transformer.disconnect_neighbour(defines.wire_type.red);
		directTransformer.transformer.connect_neighbour{wire=defines.wire_type.red, target_entity=pole.pole};
		
		if directTransformer.pole ~= pole.hash then
			--debugOutputMessage("Transformer pole hash changed!");
			makeTransformerDirty(directTransformer);
			directTransformer.pole = pole.hash;
		end
		
		if pole.transformer and directTransformer.hash ~= pole.transformer then
			directTransformer.effectiveEfficiency = pole.calcEfficiency;
			local transformer = global.transformers[pole.transformer]
			if transformer then
				transformer.effectiveEfficiency = pole.calcEfficiency;
			end
		end
	end
	
	if previousTransformerDirect and (not directTransformer or directTransformer.hash ~= previousTransformerDirect) then
		local trans = global.transformers[previousTransformerDirect];
		if trans then
			debugOutputMessage("Disconnect previous trans: 	" .. trans.hash);
	
			trans.pole = nil
			makeTransformerDirty(trans);
		end
		anyChanges = true
	end
	
	if pole.transformer and redPeerCount < 2 then
		local transfomer = getHighPoleNetworkConnectedToTransformer(pole);
		if transfomer ~= pole.transformer then
			recursiveRemoveHighPoleTransformer(pole);
		end
	end
	
	if previousPeerA ~= pole.peerA then
		local peer = global.regularPoles[previousPeerA]
		if peer then
		ebugOutputMessage("what");
			makeHighPoleDirty(peer);
		end
		anyChanges = true
	end
	if previousPeerB ~= pole.peerB then
		local peer = global.regularPoles[previousPeerB]
		if peer then
		ebugOutputMessage("what 2");
			makeHighPoleDirty(peer);
		end
		anyChanges = true
	end
	if previousTransformer ~= pole.transformer then
	--debugOutputMessage("change trans");
		anyChanges = true
	end
	if previousCalcDistance ~= pole.calcDistance then
		anyChanges = true
	end
	
	if previousCalcEfficiency ~= pole.calcEfficiency then
		anyChanges = true
	end
	
	if anyChanges then
		--debugOutputMessage("Dirtynet");
		makeHighPoleNetworkDirty(pole)
		showHighPoleIndicators(pole, 60)
	end
	
	--debugOutputMessage("Finished: " .. pole.hash .. " - " .. pole.dirtyCountDown);
end

function updateTransformer(trans)
	
	local previousPeerMajor		= trans.peerMajor;
	local previousPeerMinor		= trans.peerMinor;
	
	trans.peerMajor		= nil
	trans.peerMinor		= nil
	
	if trans.pole then
		local conPole		= global.highPoles[trans.pole];
		
		if conPole.transformer then
			if conPole.transformer > trans.hash then
				trans.peerMajor		= global.transformers[conPole.transformer];
			end
		end
	else
		if previousPeerMinor then
			previousPeerMinor.peerMajor = nil
			
			--debugOutputMessage("Test A");
			
			makeTransformerDirty(previousPeerMinor);
		elseif previousPeerMajor then
			previousPeerMajor.peerMinor = nil
			
			--debugOutputMessage("Test B");
			
			makeTransformerDirty(previousPeerMajor);
		end
	end
			
	if not trans.peerMajor then
		local minor = previousPeerMinor or trans.peerMinor
		if minor and minor.peerMajor and minor.peerMajor.hash == trans.hash then
			trans.peerMinor = minor
		end
	end
	
	if trans.peerMajor then
		--	debugOutputMessage("Test C");
		if trans.peerMajor.peerMinor ~= trans then
			makeTransformerDirty(trans.peerMajor);
			trans.peerMajor.peerMinor = trans
		end
	end
	
	if not trans.peerMajor and not trans.peerMinor then			
		if trans.arrowDownA and trans.arrowDownA.valid then
			trans.arrowDownA.destroy()
			trans.arrowDownA = nil
		end
		if trans.arrowDownB and trans.arrowDownB.valid then
			trans.arrowDownB.destroy()
			trans.arrowDownB = nil
		end
		if trans.arrowDownC and trans.arrowDownC.valid then
			trans.arrowDownC.destroy()
			trans.arrowDownC = nil
		end
		if trans.arrowUpA and trans.arrowUpA.valid then
			trans.arrowUpA.destroy()
			trans.arrowUpA = nil
		end
		if trans.arrowUpB and trans.arrowUpB.valid then
			trans.arrowUpB.destroy()
			trans.arrowUpB = nil
		end
		if trans.arrowUpC and trans.arrowUpC.valid then
			trans.arrowUpC.destroy()
			trans.arrowUpC = nil
		end
	end
end

function manageTransformer(trans)
	if trans.peerMajor then
		return
	end
	if not trans.peerMinor then
		trans.bufferSize = 10
		return
	end
	
	local	minPowerDifference	= 1000;
	local	flatEfficiency		= .975;--.15;
	local	perTickTransfer		= .5;
		
	--debugOutputMessage("Managing transformer: " .. trans.hash);
	
	local transA = trans;
	local transB = trans.peerMinor;
	
	if not transA.transformer or not transB.transformer or not transA.transformer.valid or not transB.transformer.valid then
		makeTransformerDirty(transA)
		makeTransformerDirty(transB)
		return
	end	
	
	local bufferSize = 10000;
	
	local falloffFactor		= .9985
	local growFactor		= 200
	
	transA.bufferSize 		= (transA.bufferSize or bufferSize) * falloffFactor - 10
	transB.bufferSize 		= (transB.bufferSize or bufferSize) * falloffFactor - 10
	
	transA.bufferSize		= math.max(10, transA.bufferSize)
	transB.bufferSize		= math.max(10, transB.bufferSize)
	
	transA.drainCooldown 	= (transA.drainCooldown or 0) - 1
	transB.drainCooldown 	= (transB.drainCooldown or 0) - 1
	
	if transA.interfaceOutput.energy <= transA.bufferSize / 2 and transB.interfaceDrain.energy >= transB.bufferSize then
		transB.bufferSize = transB.bufferSize + growFactor
		--addTextHelper{text="++ b; " .. transB.bufferSize, color={r=0, g=1, b=1}, position=transB.transformer.position};
	else
	end
	if transB.interfaceOutput.energy <= transB.bufferSize / 2 and transA.interfaceDrain.energy >= transA.bufferSize then
		transA.bufferSize = transA.bufferSize + growFactor
		--addTextHelper{text="++ a" .. transA.bufferSize, color={r=0, g=1, b=1}, position=transA.transformer.position};
	else
	end
	
	local effABufferSize  = transA.bufferSize
	local effBBufferSize  = transB.bufferSize
	
	--debugOutputMessage(transA.hash .. " " .. effABufferSize);
	
	transA.interfaceOutput.electric_buffer_size			= 9999999--effABufferSize
	transB.interfaceOutput.electric_buffer_size			= 9999999--effBBufferSize
	transA.interfaceDrain.electric_buffer_size			= effABufferSize
	transB.interfaceDrain.electric_buffer_size			= effBBufferSize
	
	transA.interfaceOutput.electric_input_flow_limit 	= 9999999--effABufferSize
	transB.interfaceOutput.electric_input_flow_limit 	= 9999999--effBBufferSize
	transA.interfaceDrain.electric_input_flow_limit 	= effABufferSize
	transB.interfaceDrain.electric_input_flow_limit 	= effBBufferSize
	
	transA.sendFactor 	= (transA.sendFactor or 1) * 0.99 + (global.randomGen(1e0000, 150000) / 100000) * 0.015;
	if global.randomGen(1, 50) == 1 then
		transA.sendFactor  = transA.sendFactor  * .9
	end
	
	if transA.transformer.energy == 0 then
	--	transA.drainCooldown = math.max(1, transA.drainCooldown)
	end
	if transB.transformer.energy == 0 then
	--	transB.drainCooldown = math.max(1, transB.drainCooldown)
	end
	
	if transA.drainCooldown > 0 then
		transA.interfaceDrain.electric_input_flow_limit = 0
		transA.bufferSize = 0
		--transA.interfaceOutput.energy = transA.interfaceDrain.energy
		--transA.interfaceDrain.energy = 0
	end
	if transB.drainCooldown > 0 then
		transB.interfaceDrain.electric_input_flow_limit = 0
		transB.bufferSize = 0
		--transB.interfaceOutput.energy = transB.interfaceDrain.energy
		--transB.interfaceDrain.energy = 0
	end
	
	local bufferSizeReference = 128000;
	
	local refUplinkA	= transA.bufferSize / bufferSizeReference
	local refUplinkB	= transB.bufferSize / bufferSizeReference
	
	transA.refUplink	= refUplinkA - refUplinkB
	transB.refUplink	= refUplinkB - refUplinkA
	
	fixTransformerIcon(transA)
	fixTransformerIcon(transB)
	
	local transferAtoB = transA.interfaceDrain.energy - transB.interfaceOutput.energy;
	local transferBtoA = transB.interfaceDrain.energy - transA.interfaceOutput.energy;
	
	if transferAtoB > minPowerDifference then
		local drain = transferAtoB * perTickTransfer * transA.sendFactor;
		local transfer = drain * flatEfficiency * trans.effectiveEfficiency;
		
		transA.interfaceDrain.energy			= transA.interfaceDrain.energy - drain;
		transB.interfaceOutput.energy			= transB.interfaceOutput.energy + transfer;
		transB.drainCooldown					= global.randomGen(40, 100)
		
		--addTextHelper{text="a-b add", color={r=0, g=0, b=1}, position=transB.transformer.position};
		--addTextHelper{text="a-b remove", color={r=1, g=0, b=0}, position=transA.transformer.position};
		
		--local pole = global.highPoles[transA.pole];
		--showHighPoleIndicators(pole, 60)		
	end
	if transferBtoA > minPowerDifference then
		local drain = transferBtoA * perTickTransfer * transA.sendFactor;
		local transfer = drain * flatEfficiency * trans.effectiveEfficiency;
		
		transB.interfaceDrain.energy			= transB.interfaceDrain.energy - drain;
		transA.interfaceOutput.energy			= transA.interfaceOutput.energy + transfer;
		transA.drainCooldown					= global.randomGen(40, 100)
		
		--addTextHelper{text="b-a add", color={r=0, g=0, b=1}, position=transA.transformer.position};
		--addTextHelper{text="b-a remove", color={r=1, g=0, b=0}, position=transB.transformer.position};
		
		--local pole = global.highPoles[transB.pole];
		--showHighPoleIndicators(pole, 60)		
	end
end

function fixTransformerIcon(trans)
	local position = trans.transformer.position;
	local surface = trans.transformer.surface;
		
	if trans.refUplink > 0 then
		if not trans.arrowUpA then
			trans.arrowUpA = surface.create_entity{name="highvoltage-uploading", position={x=position.x, y=position.y}, force=trans.transformer.force}
		end
	elseif trans.arrowUpA then
		if trans.arrowUpA.valid then
			trans.arrowUpA.destroy()
		end
		trans.arrowUpA = nil
	end
	if trans.refUplink > .8 then
		if not trans.arrowUpC then
			trans.arrowUpC = surface.create_entity{name="highvoltage-uploading", position={x=position.x, y=position.y-.6}, force=trans.transformer.force}
		end
	elseif trans.arrowUpC then
		if trans.arrowUpC.valid then
			trans.arrowUpC.destroy()
		end
		trans.arrowUpC = nil
	end
	if trans.refUplink > .5 then
		if not trans.arrowUpB then
			trans.arrowUpB = surface.create_entity{name="highvoltage-uploading", position={x=position.x, y=position.y-.3}, force=trans.transformer.force}
		end
	elseif trans.arrowUpB then
		if trans.arrowUpB.valid then
			trans.arrowUpB.destroy()
		end
		trans.arrowUpB = nil
	end
	if trans.refUplink < 0 then
		if not trans.arrowDownA then
			trans.arrowDownA = surface.create_entity{name="highvoltage-downloading", position={x=position.x, y=position.y}, force=trans.transformer.force}
		end
	elseif trans.arrowDownA then
		if trans.arrowDownA.valid then
			trans.arrowDownA.destroy()
		end
		trans.arrowDownA = nil
	end
	if trans.refUplink < -.5 then
		if not trans.arrowDownB then
			trans.arrowDownB = surface.create_entity{name="highvoltage-downloading", position={x=position.x, y=position.y-.3}, force=trans.transformer.force}
		end
	elseif trans.arrowDownB then
		if trans.arrowDownB.valid then
			trans.arrowDownB.destroy()
		end
		trans.arrowDownB = nil
	end
	if trans.refUplink < -.8 then
		if not trans.arrowDownC then
			trans.arrowDownC = surface.create_entity{name="highvoltage-downloading", position={x=position.x, y=position.y-.6}, force=trans.transformer.force}
		end
	elseif trans.arrowDownC then
		if trans.arrowDownC.valid then
			trans.arrowDownC.destroy()
		end
		trans.arrowDownC = nil
	end
end

function updateHighPoleInfoPanel(player, pole)
	local highPoleframe = player.gui.top.HVHighPoleFrame
	if not highPoleframe then
		highPoleframe = player.gui.top.add{type="frame",name="HVHighPoleFrame",direction="vertical"}
		highPoleframe.add{type="label",name="basicInfo",caption="Current"}
		highPoleframe.add{type="label",name="distanceInfo",caption="Current"}
		highPoleframe.add{type="progressbar",name="distanceInfoBar",direction="vertical",size=1}
	end
	
	if pole.peerA then
		if pole.peerB then
			highPoleframe.distanceInfo.caption = "2 Peers";
		else
			highPoleframe.distanceInfo.caption = "1 Peer";
		end
	elseif pole.peerB then
		highPoleframe.distanceInfo.caption = "1 Peer";
	else
		highPoleframe.distanceInfo.caption = "0 Peers";
	end
	
	if not pole.transformer then
		highPoleframe.basicInfo.caption = "No continuous connection";
	else
		local trans = global.transformers[pole.transformer];
		if trans then
			highPoleframe.basicInfo.caption = "Distance quality: " .. (math.floor(trans.effectiveEfficiency * .95 * 10000) / 100) .. "%";
		end
	end
	
	
    highPoleframe.distanceInfoBar.value = 0;
end

function removeHighPoleInfoPanel(player)
	local highPoleframe = player.gui.top.HVHighPoleFrame
	if highPoleframe then
		highPoleframe.destroy()
	end
end

function updateTransformerInfoPanel(player, trans)
	local frame = player.gui.top.HVTransformerPanel
	if not frame then
		frame = player.gui.top.add{type="frame",name="HVTransformerPanel",direction="vertical"}
		frame.add{type="label",name="transferInfo",caption="Current"}
		frame.add{type="label",name="basicInfo",caption="Current"}
		frame.add{type="label",name="chargeInfoA",caption="Current"}
		frame.add{type="progressbar",name="chargeInfoBarA",direction="vertical",size=1}
		frame.add{type="label",name="chargeInfoB",caption="Current"}
		frame.add{type="progressbar",name="chargeInfoBarB",direction="vertical",size=1}
	end
	
	local enerLocal = (trans.bufferSize or 0) / 1280;
	
	frame.chargeInfoA.caption		= "Local buffer: " .. (math.floor(enerLocal * 100) / 100) .. "%"
	frame.chargeInfoBarA.value		= enerLocal / 100;
	frame.chargeInfoB.caption		= "Not connected";
	frame.chargeInfoBarB.value		= 0;
	
	local major		= trans.peerMajor or trans;
	local minor		= trans.peerMinor or trans;
	frame.transferInfo.caption = "Transfer efficiency: " .. (math.floor(trans.effectiveEfficiency * .95 * 10000) / 100) .. "%";
	
	if not trans.pole then
		frame.basicInfo.caption = "Not connected";
	else
		local other = trans.peerMajor or trans.peerMinor;
		
		local pole = global.highPoles[minor.pole];
		
		if other then
			if pole then
				frame.basicInfo.caption 		= "Connected to " .. pole.calcDistance .. " units away."
			end

			local enerOther	= (other.bufferSize or 0) / 1280;
	
			frame.chargeInfoB.caption		= "Other buffer: " .. (math.floor(enerOther * 100) / 100) .. "%"
			frame.chargeInfoBarB.value		= enerOther / 100;	
		else
			frame.basicInfo.caption = "No continuous connection";
		end
	end
end

function removeTransformerInfoPanel(player)
	local frame = player.gui.top.HVTransformerPanel
	if frame then
		frame.destroy()
	end
end

function getHighPoleNetworkConnectedToTransformer(pole, usedList)
	local used = usedList or {};
	
	if used[pole.hash] then
		return nil
	end
	used[pole.hash] = true;
	
	if pole.transformerDirect then
		return pole.transformerDirect
	end
	
	if pole.peerA then
		local otherPole = global.highPoles[pole.peerA]
		if otherPole then
			local connected = getHighPoleNetworkConnectedToTransformer(otherPole, used)
			if connected then return connected end
		end
	end
	if pole.peerB then
		local otherPole = global.highPoles[pole.peerB]
		if otherPole then
			local connected = getHighPoleNetworkConnectedToTransformer(otherPole, used)
			if connected then return connected end
		end
	end
	
	return nil
end

function recursiveRemoveHighPoleTransformer(pole, usedList)
	local used = usedList or {};
	
	if used[pole.hash] then
		return
	end
	used[pole.hash] = true;
	
	if pole.transformer then
		pole.transformer = nil;
		makeHighPoleDirty(pole);
	end
	
	if pole.peerA then
		local otherPole = global.highPoles[pole.peerA]
		if otherPole then
			recursiveRemoveHighPoleTransformer(otherPole, used);
		end
	end
	if pole.peerB then
		local otherPole = global.highPoles[pole.peerB]
		if otherPole then
			recursiveRemoveHighPoleTransformer(otherPole, used);
		end
	end
end

function showHighPoleIndicators(pole, duration, usedList)
	local used = usedList or {};

	if used[pole.hash] then
		return
	end
	used[pole.hash] = true;
	
	if pole.peerA then
		local otherPole = global.highPoles[pole.peerA]
		if otherPole and not used[otherPole.hash] then
			if otherPole.transformer then
				addBeamHelper("highvoltage-connection-indication-powerbox", pole.pole, otherPole.pole, duration);
			else
				addBeamHelper("highvoltage-connection-indication-bad", pole.pole, otherPole.pole, duration);
			end
			showHighPoleIndicators(otherPole, duration, used);
		end
	end
	if pole.peerB then
		local otherPole = global.highPoles[pole.peerB]
		if otherPole then -- and not used[otherPole.hash]
			if otherPole.transformer then
				addBeamHelper("highvoltage-connection-indication-powerbox", pole.pole, otherPole.pole, duration);
			else
				addBeamHelper("highvoltage-connection-indication-bad", pole.pole, otherPole.pole, duration);
			end
			showHighPoleIndicators(otherPole, duration, used);
		end
	end
	if pole.transformerDirect then
		local trans = global.transformers[pole.transformerDirect]
		if trans then
			if trans.enabled then
				addBeamHelper("highvoltage-connection-indication-powerbox", pole.pole, trans.transformer, duration);
			end
		end
	end
end

function showTransformerIndicators(trans, duration, usedList)
	local used = usedList or {};

	if used[trans.hash] then
		return
	end
	used[trans.hash] = true;
	
	if trans.pole then
		local otherPole = global.highPoles[trans.pole]
		if otherPole and not used[otherPole.hash] then
			addBeamHelper("highvoltage-connection-indication-powerbox", trans.transformer, otherPole.pole, duration);
			showHighPoleIndicators(otherPole, duration, used);
		end
	end
end