function addPowerBox(entity, prototype)
	local existingBox = findPowerBoxByEntity(entity);
	
	if existingBox then
		makePowerBoxNetworkDirty(existingPole);
		return;
	end
	
    local object = {
       box = entity,
	   arrow = nil,
	   hash = hash_entity(entity),
	   dirtyCountDown = 0,
	   regularPoles = {},
	   regularPoleCount = 0,
	   enabled = true
    }
	
	global.powerBoxes[object.hash] = object;
	
	local box = global.powerBoxes[object.hash];
	
	makePowerBoxDirty(box);
	return box;
end

function removePowerBoxData(box)	
	if box.arrow and box.arrow.valid then
		box.arrow.destroy()
	end
	
	makePowerBoxNetworkDirty(box)
	
	global.powerBoxes[box.hash] = nil
end

function updatePowerBox(box)
	if not box.box.valid then
		removePowerBoxData(box)
		return
	end
	
	local electricConnected = box.box.neighbours["copper"];
	local previousRegularPoles = box.regularPoles or {};
	local previousRegularPoleCount = box.regularPoleCount;
	local anyChanges = false
	
	box.regularPoles = {};
	box.regularPoleCount = 0;
	
	for k,v in pairs(electricConnected) do
		local otherPole = findRegularPoleByEntity(v);
		
		if otherPole then
			box.regularPoles[k] = otherPole.hash;
			box.regularPoleCount = box.regularPoleCount + 1;
		else
			local powerBox = findPowerBoxByEntity(v);
			
			if powerBox then
				box.box.disconnect_neighbour(v);
			end
		end
	end
	
	if box.regularPoleCount ~= previousRegularPoleCount then
		for k,v in pairs(previousRegularPoles) do
			local pole = global.regularPoles[v]
			if pole then
				makeRegularPoleDirty(pole);
			end
		end
		anyChanges = true
	end
	
	if anyChanges then
		makePowerBoxNetworkDirty(box)
		showOutwardPowerBoxIndicators(box, 60);
	end
end

function showOutwardPowerBoxIndicators(box, duration)
	for k,v in pairs(box.regularPoles) do
		local otherPole = global.regularPoles[v]
		if otherPole then
			showInwardRegularPoleIndicators(otherPole, duration, {});
			showOutwardRegularPoleIndicators(otherPole, duration, {});
		end
	end
end