function addRegularPole(entity, prototype)
	local existingPole = findRegularPoleByEntity(entity);
	
	if existingPole then
		makeRegularPoleNetworkDirty(existingPole);
		return;
	end
	
    local object = {
       pole = entity,
	   arrow = nil,
	   hash = hash_entity(entity),
	   calcDistance = 99,
	   dirtyCountDown = 0,
	   peers = {},
	   peerCount = 0,
	   enabled = true,
	   mainProvider = nil,
	   powerBox = nil,
	   powerBoxNetwork = nil,
	   prototypeDistanceMitigation = prototype.quality,
	   maximumDistance = prototype.maximumDistance,
	   effectiveDistanceMitigation = 1
    }
	
	global.regularPoles[object.hash] = object;
	makeRegularPoleDirty(global.regularPoles[object.hash]);
	
	return global.regularPoles[object.hash];
end

function removeRegularPoleData(pole)
--	debugOutputMessage("Removing pole data: " .. pole.hash);
	
	for k,v in pairs(pole.peers) do
		local peer = global.regularPoles[v]
		if peer then	
			peer.peers = {}
		end
	end
	
	if pole.arrow and pole.arrow.valid then
		pole.arrow.destroy()
	end
	
	makeRegularPoleNetworkDirty(pole)
	
	global.regularPoles[pole.hash] = nil
end

function replaceRegularPoleEntity(pole)
	local currentDisabled = (string.sub(pole.pole.name,-string.len("-disabled"))=="-disabled")
	
--	debugOutputMessage("Replacing: " .. pole.hash);
			
	if currentDisabled == pole.enabled then	
		local entityNameRegular = pole.pole.name
		
		if currentDisabled then
			entityNameRegular = entityNameRegular:sub(1, -10);
		end
	
		local surface = pole.pole.surface;
		local newPole;
		
		local oldNeighbours = table.deepcopy(pole.pole.neighbours);
			
		if pole.enabled then
		--	addTextHelper({position=pole.pole.position, text="Enabled", color={r=0.5, g=1.0, b=.5}})
		--	debugOutputMessage("Enabled: " .. pole.hash);
			newPole = surface.create_entity{name=entityNameRegular, position=pole.pole.position, force=pole.pole.force}
			if pole.arrow then
				pole.arrow.destroy()
			end
		else
		--	addTextHelper({position=pole.pole.position, text="Disabled", color={r=1.0, g=0.5, g=.5}})
		--	debugOutputMessage("Disabled: " .. pole.hash);
			newPole = surface.create_entity{name=entityNameRegular .. "-disabled", position=pole.pole.position, force=pole.pole.force}
			pole.arrow = surface.create_entity{name="highvoltage-disabled-warning", position=pole.pole.position, force=pole.pole.force}
		end
		
		pole.pole.destroy()
		pole.pole = newPole
		
		pole.pole.disconnect_neighbour();
		
		for connectType,neighbourGroup in pairs(oldNeighbours) do
			if connectType == "copper" then
				for k,v in pairs(neighbourGroup) do
					pole.pole.connect_neighbour(v);
				end
			elseif connectType == "red" then
				for k,v in pairs(neighbourGroup) do
					pole.pole.connect_neighbour({wire = defines.wire_type.red, target_entity = v});
				end
			elseif connectType == "green" then
				for k,v in pairs(neighbourGroup) do
					pole.pole.connect_neighbour({wire = defines.wire_type.green, target_entity = v});
				end
			end
		end
	end
end

function updateRegularPole(pole)
	if not pole.pole.valid then
		removeRegularPoleData(pole)
		return
	end

	local electricConnected = pole.pole.neighbours["copper"];
	local smallestProviderDistance = pole.calcDistance;
	local previousPeers = pole.peers
	local previousPeerCount = pole.peerCount
	local previousDistanceMitigation = pole.effectiveDistanceMitigation
	local previousEnabled = pole.enabled
	local previousMainProvider = pole.mainProvider
	local previousPowerBox = pole.powerBox
	local previousPowerBoxNetwork = pole.powerBoxNetwork
	local anyChanges = false
	
	pole.peers = {};
	pole.peerCount = 0;
	pole.powerBox = nil;
	pole.powerBoxNetwork = nil;
	pole.mainProvider = nil;
	
	pole.effectiveDistanceMitigation = math.max(.1, pole.prototypeDistanceMitigation);
	
	smallestProviderDistance = 99
	
	for k,v in pairs(electricConnected) do
		local otherPole = findRegularPoleByEntity(v);
		
		if otherPole then
			pole.peers[k] = otherPole.hash;
			pole.peerCount = pole.peerCount + 1;
			
			local localDist = gameLogicDistance(pole.pole.position.x, pole.pole.position.y, v.position.x, v.position.y);
			localDist = localDist / pole.effectiveDistanceMitigation;
			
			local otherProviderDistance = otherPole.calcDistance + localDist;
			
			if otherProviderDistance < smallestProviderDistance then
				smallestProviderDistance = otherProviderDistance
				pole.mainProvider = otherPole.hash;
				pole.powerBoxNetwork = otherPole.powerBoxNetwork;
			end
		else
			local powerBox = findPowerBoxByEntity(v);
			
			if powerBox then
				pole.powerBox = powerBox.hash;
				pole.powerBoxNetwork = powerBox.hash;
				
				local localDist = gameLogicDistance(pole.pole.position.x, pole.pole.position.y, v.position.x, v.position.y);
				localDist = localDist / pole.effectiveDistanceMitigation;
				
				smallestProviderDistance = localDist
			else
				addTextHelper{position=pole.pole.position, text="(X)"}
				addTextHelper{position=v.position, text="(X)"}
				pole.pole.disconnect_neighbour(v);
			end
		end
	end
	
	if smallestProviderDistance > 999 then
		smallestProviderDistance = 999
	end
	
	if smallestProviderDistance ~= pole.calcDistance then
		pole.calcDistance = smallestProviderDistance
		anyChanges = true
	end
	
	if pole.calcDistance <= pole.maximumDistance then
		pole.enabled = true
	else
		pole.enabled = false
	end
	
	if pole.enabled ~= previousEnabled then
		replaceRegularPoleEntity(pole);
		anyChanges = true
	end
	
	if previousQuality ~= pole.effectiveQuality then
		anyChanges = true
	end
	
	if pole.peerCount ~= previousPeerCount then
		for k,v in pairs(previousPeers) do
			local peer = global.regularPoles[v]
			if peer then
				makeRegularPoleDirty(peer);
			end
		end
		anyChanges = true
	end
	
	if pole.powerBox ~= previousPowerBox then
		local oldPowerBox = global.powerBoxes[previousPowerBox];
		local newPowerBox = global.powerBoxes[pole.powerBox];
		if oldPowerBox then
			makePowerBoxDirty(oldPowerBox);
		end
		if newPowerBox then
			makePowerBoxDirty(newPowerBox);
		end
	end
	
	if pole.powerBoxNetwork ~= previousPowerBoxNetwork then
		local oldPowerBox = global.powerBoxes[previousPowerBoxNetwork];
		local newPowerBox = global.powerBoxes[pole.powerBoxNetwork];
		if oldPowerBox then
			makePowerBoxDirty(oldPowerBox);
		end
		if newPowerBox then
			makePowerBoxDirty(newPowerBox);
		end
	end
	
	if anyChanges then
		makeRegularPoleNetworkDirty(pole)
		showInwardRegularPoleIndicators(pole, 60);
		showOutwardRegularPoleIndicators(pole, 60);
		isolateRegularPoleToPowerBoxNetwork(pole, pole.powerBoxNetwork);
	end
	
	if pole.dirtyCountDown == 0 then
	--	debugOutputMessage("Cleaned: " .. pole.hash);
	end
end

function isolateRegularPoleToPowerBoxNetwork(pole, boxHash)
	if not pole.powerBoxNetwork then
		return
	end

	electricConnected = pole.pole.neighbours["copper"];

	local anyChanges = false;
	
	for k,v in pairs(electricConnected) do
		local otherPole = findRegularPoleByEntity(v);
		
		if otherPole then
			if otherPole.powerBoxNetwork then
				if otherPole.powerBoxNetwork ~= boxHash then
					addTextHelper{position=pole.pole.position, text="(X)"}
					addTextHelper{position=otherPole.pole.position, text="(X)"}
					pole.pole.disconnect_neighbour(otherPole.pole);
					anyChanges = true;
				end
			end
		else
			local powerBox = findPowerBoxByEntity(v);
			
			if powerBox then
				if powerBox.hash ~= boxHash then
					addTextHelper{position=pole.pole.position, text="(X)"}
					addTextHelper{position=powerBox.box.position, text="(X)"}
					pole.pole.disconnect_neighbour(powerBox.box);
					anyChanges = true;
				end
			end
		end
	end
	
	if anyChanges then
		makeRegularPoleNetworkDirty(pole);
	end
end

function updateRegularPoleInfoPanel(player, pole)
	local energy_frame = player.gui.top.energy_frame
	if not energy_frame then
		energy_frame = player.gui.top.add{type="frame",name="energy_frame",direction="vertical"}
		energy_frame.add{type="label",name="basicInfo",caption="Current"}
		energy_frame.add{type="label",name="distanceInfo",caption="Current"}
		energy_frame.add{type="progressbar",name="distanceInfoBar",direction="vertical",size=1}
	end
	
    energy_frame.basicInfo.caption = "Distance quality: " .. pole.effectiveDistanceMitigation;
	energy_frame.basicInfo.caption = energy_frame.basicInfo.caption .. " - Peers: " .. pole.peerCount;
	
	if pole.calcDistance < 999 then
		local calcDistShow = math.ceil(pole.calcDistance * 2) / 2
		energy_frame.distanceInfo.caption = "Distance to provider: " .. calcDistShow .. "/" .. pole.maximumDistance;
	else
		energy_frame.distanceInfo.caption = "Not provided for!"
	end
	
	if pole.enabled then
		energy_frame.distanceInfo.caption = energy_frame.distanceInfo.caption .. " (enabled)";
	else
		energy_frame.distanceInfo.caption = energy_frame.distanceInfo.caption .. " (disabled)";
	end
	
    energy_frame.distanceInfoBar.value = math.max(0, pole.maximumDistance - pole.calcDistance) / math.max(1, pole.maximumDistance);
end

function removeRegularPoleInfoPanel(player)
	local energy_frame = player.gui.top.energy_frame
	if energy_frame then
		energy_frame.destroy()
	end
end

function showInwardRegularPoleIndicators(pole, duration, usedList)
	local used = usedList or {};

	if used[pole.hash] then
		return
	end
	used[pole.hash] = true;
		
	if pole.mainProvider then
		local otherPole = global.regularPoles[pole.mainProvider];
		if not otherPole then
			return
		end
		
		if pole.enabled then
			addBeamHelper("highvoltage-connection-indication-good", pole.pole, otherPole.pole, duration);
		else
			addBeamHelper("highvoltage-connection-indication-bad", pole.pole, otherPole.pole, duration);
		end
		showInwardRegularPoleIndicators(otherPole, duration, used);
	end
	
	if pole.powerBox then
		local box = global.powerBoxes[pole.powerBox]
			
		if box then
			addBeamHelper("highvoltage-connection-indication-powerbox", box.box, pole.pole, duration);
		end
	end
	
	--if pole.powerBoxNetwork then
	--	local box = global.powerBoxes[pole.powerBoxNetwork]
	--		
	--	if box then
	--		addBeamHelper("highvoltage-connection-indication-powerbox", box.box, pole.pole, duration);
	--	end
	--end
end

function showOutwardRegularPoleIndicators(pole, duration, usedList)
	local used = usedList or {};

	if used[pole.hash] then
		return
	end
	used[pole.hash] = true;
	
	for k,v in pairs(pole.peers) do
		local otherPole = global.regularPoles[v]
		if otherPole and otherPole.mainProvider == pole.hash then
			if otherPole.enabled then
				addBeamHelper("highvoltage-connection-indication-good", pole.pole, otherPole.pole, duration);
			else
				addBeamHelper("highvoltage-connection-indication-bad", pole.pole, otherPole.pole, duration);
			end
			showOutwardRegularPoleIndicators(otherPole, duration, used);
		end
	end
end